""" 
-----------------------
The colors module
-----------------------

Provide a few color conversion facilities.

The reference color format here is the one of a numpy array
with shape (3,) and dtype np.uint8. Beware that this is not the same 
convention as in matplotlib.colors. In this package, interacting with 
this package, it is recommended to always prefer our colors.color_to_rgb
to matplotlib.colors.to_rgb.


Otherwise, you could run in incompatibily issues.


"""


from typeguard import typechecked
import numpy as np
from matplotlib import colors



# Below, the colors are supposed to be given as rgb triples.
# or that matplotlib.colors can interprete them as colors,
# that is, the colors.to_rgb function accepts them as input.


@typechecked
def color_to_rgb(
    color: tuple[int, int, int]
    | list[int]
    | tuple[float, float, float]
    | list[float]
    | str
    | np.ndarray
) -> np.ndarray:
    """Convert matplotlib color to 8-bits rgb color.

    Args:
        color (tuple[int, int, int] | list[int] | \
            tuple[float, float, float] |\
            list[float] | str | np.ndarray):an 8-bits rgb color or its float \
            rescaled version, with values in [0,1] or a string recognized \
            by matplotlib.to_rgb.

    Raises:
        ValueError: "Expecting a 3-channels color (rgb) or a string."
        ValueError: "To specify your color by a triple of integers,
            they should belong to the closed interval [0,255]."

    Returns:
        np.ndarray: an np.ndarray of shape (3,) and dtype 'uint8', representing
            the rgb color.
    """
    if isinstance(color, str):
        color = colors.to_rgb(color)
    if isinstance(color, np.ndarray):
        color=list(map(int,list(color)))
    if len(color) != 3:
        raise ValueError("Expecting a 3-channels color (rgb) or a string.")
    if isinstance(color[0], float):
        return np.round(255 * np.array(color)).astype("uint8")

    if isinstance(color[0], int):
        if -1 < color[0] < 256 and -1 < color[1] < 256 and -1 < color[2] < 256:
            return np.array(color)
        else:
            raise ValueError(
                "To specify your color by a triple of integers, "
                + "they should belong to the closed interval [0,255]."
            )



def binary2grayscale(image: np.ndarray) -> np.ndarray:
    """Convert a binary image to rgb image.

    Args:
        image (np.ndarray): A binary image as a 2d np.ndarray

    Returns:
        np.ndarray: the corresponding grayscale image.
    """
    return 255 * image.astype("uint8")


def grayscale2rgb(image: np.ndarray) -> np.ndarray:
    """Convert a grayscale image to rgb image.

    Args:
        image (np.ndarray): A grayscale image as a 2d np.ndarray with
            dtype 'uint8'.

    Returns:
        np.ndarray: the corresponding rgb image.
    """
    return np.stack([image] * 3, axis=2)


def binary2rgb(image: np.ndarray) -> np.ndarray:
    """Convert a binary image to rgb image.

    Args:
        image (np.ndarray): A binary image as a 2d np.ndarray

    Returns:
        np.ndarray: the corresponding rgb image
    """
    return grayscale2rgb(binary2grayscale(image))


def to_rgb(image: np.ndarray) -> np.ndarray:
    """Convert image to the rgb format.

    Args:
        image (np.ndarray): The image to be converted.

    Raises:
        ValueError: The expected dtype of the ndarray is either 'bool'
            or 'uint8'. The expected shape is either 2d or 3d with
            shape[2]==3.

    Returns:
        np.ndarray: The image converted to the rgb format.
    """
    if image.dtype == np.dtype(bool):
        return binary2rgb(image)
    if image.dtype == np.dtype("uint8") and len(image.shape) == 2:
        return grayscale2rgb(image)
    if (
        image.dtype == np.dtype("uint8")
        and len(image.shape) == 3
        and image.shape[2] == 3
    ):
        return image
    else:
        raise ValueError(
            "The expected dtype of the ndarray is either 'bool' or 'uint8'."
            + "The expected shape is either 2d or 3d with shape[2]==3."
        )



"""
--------------------------------
The image_annotation module
--------------------------------



    This module deals with displaying text on a Numpy image.

"""

import skimage
import numpy as np
import warnings
import os
import csv

from . import colors


absolute_path = os.path.dirname(__file__)


png_alphabet_folder = os.path.join(
    absolute_path, "alphabet", "liberation_serif"
)
path_to_csv = os.path.join(png_alphabet_folder, "metadata.csv")
characters_as_images = {}
with open(path_to_csv, "r", newline="") as csvfile:
    reader = csv.reader(csvfile, dialect="excel")
    for row in reader:
        png_path = os.path.join(png_alphabet_folder, row[0])
        png_char = skimage.io.imread(png_path, as_gray=True).astype("bool")
        png_char ^= True
        character = row[1]
        characters_as_images[character] = png_char

characters_as_images[" "] = np.ones(
    characters_as_images["e"].shape, dtype="bool"
)


# This will be convenient to have all characters of the exact same
#  height and to have a small white margin on the left of
#  all characters.

max_height = max(val.shape[0] for val in characters_as_images.values())
max_width = max(val.shape[1] for val in characters_as_images.values())

for c in characters_as_images:
    shape_ = characters_as_images[c].shape
    diff = max_height - shape_[0]
    top = diff // 2
    bottom = diff - top
    characters_as_images[c] = np.vstack(
        [
            np.ones((top, shape_[1]), dtype="bool"),
            characters_as_images[c],
            np.ones((bottom, shape_[1]), dtype="bool"),
        ]
    )
    characters_as_images[c] = np.hstack(
        [
            characters_as_images[c],
            np.ones((max_height, max_width // 10 + 1), dtype="bool"),
        ]
    )


def string_to_image(s: str, relative_line_spacing: float = 0.1) -> np.ndarray:
    """Produce image of the given string.

    Args:
        s (str): The string that must appear on the image.
        relative_line_spacing (float, optional): The line spacing, relative
            to the height of ONE line. Defaults to 0.1.

    Returns:
        np.ndarray: boolean image displaying the input string. The text
        corresponding to null pixels.
    """
    if not s:
        print("The input string is empty, returning None.")
        return None
    else:
        lines = s.split("\n")
        image_lines = []
        max_width = 0

        for line in lines:
            image_lines.append(
                np.hstack([characters_as_images[c] for c in line])
            )
            max_width = max(max_width, image_lines[-1].shape[1])

        for i in range(len(image_lines)):
            line = image_lines[i]
            height, width = line.shape
            line_spacing = int(height * relative_line_spacing)
            image_lines[i] = np.hstack(
                [line, np.ones((height, max_width - width), dtype="bool")]
            )

            if i != len(image_lines) - 1:
                line = image_lines[i]
                image_lines[i] = np.vstack(
                    [line, np.ones((line_spacing, max_width), dtype="bool")]
                )

        return np.vstack(image_lines)



# currently unused function
def _annotated_image(
    image,
    text,
    x,
    y,
    size=None,
    relative_size=0.1,
    color=(0, 0, 0),
    background_color=(255, 255, 255),
    relative_padding=0.05,
):
    text_image = (1 - string_to_image(text)).astype("bool")

    # We parse colors
    color = colors.color_to_rgb(color)
    background_color = colors.color_to_rgb(background_color)


    # We deal with size issues
    position_issue = (
        "The required position for your text ("
        + text
        + ")is outside the image or too close to its border, "
        + "we push the text inside the image and scale it to fit in."
    )
    size_issue = (
        "Your text ("
        + text
        + ") is too high or too wide to fit in the image at the required "
        + "position, we adapt its size to fit in."
    )

    if size is not None:
        relative_size = size / image.shape[0]

    has_position_issue = False
    if y > image.shape[0] - 7:
        y = max(0, image.shape[0] - 7)
        has_position_issue = True
    if y < 7:
        y = min(7, image.shape[0])
        has_position_issue = True

    if x > image.shape[1] - 7:
        x = max(0, image.shape[1] - 7)
        has_position_issue = True
        if x < 7:
            x = min(7, image.shape)
            has_position_issue = True
    if has_position_issue:
        warnings.warn(position_issue)
    has_size_issue = False
    if y + image.shape[0] * relative_size > image.shape[0]:
        relative_size = 0.9 * (image.shape[0] - y) / image.shape[0]
        has_size_issue = True

    text_height = image.shape[0] * relative_size
    scaling_factor = text_height / text_image.shape[0]
    text_image = skimage.transform.rescale(text_image, scaling_factor)

    if text_image.shape[1] + x > image.shape[1]:
        scaling_factor = 0.9 * (image.shape[1] - x) / text_image.shape[1]
        text_image = skimage.transform.rescale(text_image, scaling_factor)
        has_size_issue = True

    if has_size_issue:
        warnings.warn(size_issue)

    height, width = text_image.shape

    padding = int(height * relative_padding)

    if len(image.shape) != 3:
        new_image = skimage.color.gray2rgb(image)
    else:
        new_image = image.copy()

    text_mask = np.zeros(new_image.shape[:2], dtype="bool")
    text_mask[y : y + height, x : x + width] = text_image

    if background_color is not None:
        new_image[
            y - padding : y + height + padding,
            x - padding : x + width + padding,
        ] = np.array(background_color)

    new_image[text_mask] = np.array(color)

    return new_image


def center_text(
    text: str, shape: tuple[int, int], color="black", background_color="white"
):
    """Center given text over a monochrome background to create a np.ndarray.

    Args:
        text (str): The text to print
        shape (tuple[int,int]): The shape of the background
        color: The color of the background, of a type
            acceptable by colors.color_to_rgb. Defaults to black.
        background_color: The color of the background, of a type
            acceptable by colors.color_to_rgb. Defaults to white.

    Returns:
        np.ndarray: The resulting image as an rgb np.ndarray of dtype 'uint8'.
            The text fills 90% of the width or of the height of the input shape,
            depending on its own proportions.
    """
    if text == "":
        text = " "

    # We parse colors
    color = colors.color_to_rgb(color)
    background_color = colors.color_to_rgb(background_color)

    img = np.array([[background_color] * shape[1]] * shape[0])
    text_zone = string_to_image(text) ^ True
    supp = np.where(text_zone == 0)
    slice_x = slice(np.amin(supp[1]), np.amax(supp[1]))
    slice_y = slice(np.amin(supp[0]), np.amax(supp[0]))
    text_zone = text_zone[slice_y, slice_x]

    acceptable_width = (9 * shape[1]) / 10
    acceptable_height = (9 * shape[0]) / 10

    scaling_factor = min(
        acceptable_height / text_zone.shape[0],
        acceptable_width / text_zone.shape[1],
    )

    text_zone = skimage.transform.rescale(text_zone, scaling_factor)
    h = text_zone.shape[0]
    w = text_zone.shape[1]
    x = (shape[1] - w) // 2
    y = (shape[0] - h) // 2
    text_mask = np.zeros(shape[:2], dtype="bool")
    text_mask[y : y + h, x : x + w] = text_zone

    img[text_mask] = color

    return img


# For testing purposes
# from matplotlib import pyplot as plt


# def plot_page(page, cmap="gray_r"):
#     plt.figure(figsize=(20, 20))
#     plt.imshow(page, cmap=cmap)
#     plt.show()


# img = center_text(
#     "ah, ah ah,\n Oh où vas Tu? Não quer ficar?",
#     (400, 700),
#     color="purple",
#     background_color="orange",
# )
# plot_page(img)
# img = center_text(
#     "ah, ah ah,\n Oh où vas Tu? Não quer ficar?",
#     (400, 700),
#     color="purple",
#     background_color=[20,90,165],
# )
# plot_page(img)



"""
--------------------------------
The np_clickable_image module
--------------------------------

    This module implements ClickableImage and some subclasses.

"""



from __future__ import annotations

import numpy as np
from typing import Callable, Any
from matplotlib import pyplot as plt
import skimage
import os
import sys
import inspect, warnings


from . import image_annotation
from . import colors




class _CombinableFunc:
    """A class of callables mimicking positional keywords functions.

    A class of callables mimicking functions with only positional
    keyword arguments with better control on signature and default
    values. This class does not care about the specific order of the
    positional keyword arguments.


    """

    @staticmethod
    def defaults_dictionary(f: Callable[Any, Any]) -> dict:
        """Parse the keyword(-only) arguments default values.

        Args:
            f (Callable[Any,Any]): A function.

        Returns:
            dict: A dictionary with keys given by the keyword arguments
                and keyword-only arguments of f that maps these keys to
                the corresponding default values.
        """
        data = inspect.getfullargspec(f)
        args = data.args
        defaults = data.defaults
        n = len(defaults)
        k = len(args) - n
        if k != 0:
            raise TypeError(
                "There are some positional non-keyword "
                + "arguments in your function"
            )
        args_dic = {args[k + i]: defaults[i] for i in range(n)}
        kwonlydefaults = data.kwonlydefaults
        if kwonlydefaults is None:
            kwonlydefaults = {}
        return args_dic | kwonlydefaults

    def __init__(self, func: Callable, defaults_dic: dict = None):
        """Build the callable.

            Build the callable from a  keyword(-only) function and
            optionally a dictionary. The function can be defined with
            a signature ( \*\*kwargs, \*, ...). In this case, the placeholder
            for a detailed specification of keyword
            arguments is the dictionary, see below. If the dictionary
            is not passed, the function's signature must specify every
            admissible positional keyword argument. Functions using
            also keyword-only arguments can be passed but the
            keyword-only arguments with no default value will be
            ignored. If a dictionary is passed, the function's signature
            is ignored.


        Args:
            func (Callable): A function with solely keyword and
                keyword-only arguments;
            dic (dict, optional): A dictionary specifying all admissible
                positional keyword arguments TOGETHER WITH their
                default values. Defaults to None.
        """

        if defaults_dic is None:
            defaults_dic = _CombinableFunc.defaults_dictionary(func)

        self.defaults_dic = defaults_dic
        self.function = func

    def __call__(self, **kwargs):
        for key in kwargs.keys():
            if key not in self.defaults_dic:
                raise KeyError(
                    "The argument " + key + " is not valid for this callable."
                )
        completed_args = kwargs | {
            key: self.defaults_dic[key]
            for key in self.defaults_dic
            if key not in kwargs
        }
        return self.function(**completed_args)


# # Testing _CombinableFunc
# def _f(a=2, *, u=3):
#     return a + u


# f = _CombinableFunc(_f)
# print(f())  # 5
# print(f(a=7))  # 10
# print(f(a=7, u=12))  # 19
# print(f(u=13))  # 15


# def _g(**kwargs):
#     return sum([*kwargs.values()])


# defaults = {"a": 2, "u": 3}
# g = _CombinableFunc(_g, defaults)
# print(g())  # 5
# print(g(a=7))  # 10
# print(g(a=7, u=12))  # 19
# print(g(u=13))  # 15


# A decorator to define _CombinableFuncs as functions with detailed
# signatures.
def _combinable_func(g):
    return _CombinableFunc(g)


# @_combinable_func
# def f(a=2, *, u=3):
#     return a + u


# print(f())  # 5
# print(f(a=7))  # 10
# print(f(a=7, u=12))  # 19
# print(f(u=13))  # 15
# print(f.defaults_dic)







class ClickableImage:
    """A class for clickable images, made with matplotlib and numpy."""

    def __init__(
        self,
        image: Callable[Any, np.ndarray] | np.ndarray,
        shape: tuple[int, ...],
        regions: list[np.ndarray],
        callbacks: list[Callable[dict, None]],
        vars_dic: dict,
    ):
        """Builder method

        Args:
            image (Callable[Any, np.ndarray] | np.ndarray): A function that
                returns an image or a 'constant' image. If a function is
                passed, it must have only keyword and keyword-only arguments.
                The keyword-only arguments with no default value will be
                ignored.

            shape (tuple[int, ...]): shape of the ndarray returned by
                image
            regions (list[np.ndarray]): boolean arrays A with A.shape=shape[:2]

            callbacks (list[Callable[dict, None]]): functions to be
                called when the regions are clicked
                (the ith callback belongs to the ith region)
            vars_dic (dict): some variable values stored in a dict. They
                are meant to define the behaviour of the image attributes
                of the current instance and possibly other ones, in view of
                interactions between instances (stacking).

        Raises:
            ValueError: The number of regions does not equal the
                number of callbacks.
            ValueError: The regions' shapes do not all equal shape[:2]
            ValueError: The passed regions do not have
                dtype('bool') dtype.
        """
        # Compatibility checks.
        if not len(callbacks) == len(regions):
            raise ValueError(
                "The number of regions does not equal the"
                + " number of callbacks"
            )
        if not all([region.shape == shape[:2] for region in regions]):
            raise ValueError("The regions' shapes do not all equal shape[:2]")
        if not all([region.dtype == np.dtype("bool") for region in regions]):
            raise ValueError(
                "The passed regions do not have dtype('bool') dtype."
            )
        if isinstance(image, _CombinableFunc):
            self.image = image
        elif isinstance(image, np.ndarray):
            self.image = _CombinableFunc(lambda: image, {})
        else:
            self.image = _CombinableFunc(image)
        self.shape = shape
        self.regions = regions
        self.callbacks = callbacks
        self.vars_dic = vars_dic

    def get_image(self, dic: dict | None = None) -> np.ndarray:
        if dic is None:
            dic = self.vars_dic
        keys = self.image.defaults_dic.keys()
        kwargs = {key: dic[key] for key in keys if key in dic}
        output = self.image(**kwargs)
        if output.shape != self.shape:
            raise ValueError(
                "The image attribute of your ClickableImage "
                + "did not return a ndarray with shape equal to the shape "
                + "attribute."
            )

        return output

    def click_action(self, index: int):
        callback = self.callbacks[index]
        callback(self.vars_dic)

    def use(
        self, return_vars: list[str] | None = None, **plot_options
    ) -> dict:
        """Use the clickable image to get some user input.

        For this method to work, all the keyword arguments for image
        must be contained in vars_dic.

        Args:
            return_vars (list[str] | None, optional):
                The keys of vars_dic we are interested in. Defaults to None.
                If None is received, all vars_dic will be returned after
                the interaction.

            **plot_options: the optional arguments to be passed to
                imshow. Useful to affect the display.

        Returns:
            dict: The requested dictionary of variables values.
        """
        # Display initialization.
        fig, ax = plt.subplots()
        fig.set_size_inches(10, 5)
        ax.set_axis_off()
        ax.imshow(self.get_image(), **plot_options)

        # Event handling.
        def onclick(event):
            if ax == event.inaxes:
                y = int(event.ydata + 0.5)
                x = int(event.xdata + 0.5)

                # Variables and display update.
                for i, region in enumerate(self.regions):
                    if region[y, x]:
                        self.click_action(i)
            ax.imshow(self.get_image(), **plot_options)
            plt.pause(0.0001)

        cid = fig.canvas.mpl_connect("button_press_event", onclick)
        plt.show()

        if return_vars is None:
            return self.vars_dic
        return {key: self.vars_dic[key] for key in return_vars}

    def has_interaction(self, clickable: list[ClickableImage]):
        d1 = self.vars_dic | self.image.defaults_dic
        d2 = clickable.vars_dic | clickable.image.defaults_dic
        keys1 = set(d1.keys())
        keys2 = set(d2.keys())
        return bool(keys1.intersection(keys2))

    def transpose(self):
        def new_image_func(**kwargs):
            to_be_transposed = self.image.function(**kwargs)
            ndim = to_be_transposed.ndim
            return np.transpose(
                to_be_transposed, axes=(1, 0) + tuple(range(2, ndim))
            )

        new_image = _CombinableFunc(new_image_func, self.image.defaults_dic)
        new_regions = [np.transpose(region) for region in self.regions]
        new_shape = (self.shape[1], self.shape[0]) + self.shape[2:]

        return ClickableImage(
            new_image, new_shape, new_regions, self.callbacks, self.vars_dic
        )

    @staticmethod
    def constant_to_clickable(img: np.ndarray) -> ClickableImage:
        """Convert np.ndarray image to ClickableImage.

        Args:
            img (np.ndarray): an image

        Returns:
            ClickableImage: The corresponding "constant" ClickableImage.
        """
        return ClickableImage(img, img.shape, [], [], {})

    @staticmethod
    def hstack(
        clickables: list[ClickableImage | np.ndarray],
    ) -> ClickableImage:
        """Stack horizontally the input ClickableImages or np.ndarray images.

        Args:
            clickables (list[ClickableImage|np.ndarray]): list of
                ClickableImages or images, in the second case they are
                treated as 'constant' ClickableImages.

        Raises:
            ValueError: The clickables are not stackable as demanded, due
                to shape issues.

        Returns:
            ClickableImage: The ClickableImage obtained by stacking
                horizontally the input clickables/images, with possible
                interactions if they share keys of their
                image.defaults_dic | vars_dic
        """
        height = clickables[0].shape[0]

        for i in range(len(clickables)):
            if isinstance(clickables[i], np.ndarray):
                clickables[i] = ClickableImage.constant_to_clickable(
                    clickables[i]
                )

        def measure_depth(clickable: np.ndarray) -> int | None:
            A = clickable.get_image()
            return None if A.ndim == 2 else A.shape[2]

        depth = measure_depth(clickables[0])
        same_heights = all(
            [clickable.shape[0] == height for clickable in clickables]
        )

        same_depth = all(
            [measure_depth(clickable) == depth for clickable in clickables]
        )

        if not (same_heights and same_depth):
            raise ValueError(
                "The clickables are not stackable as demanded,"
                + " due to shape issues."
                + " Remember this could include a depth issue."
            )
        n_blocks = len(clickables)
        block_x_coords = [
            sum(clickables[i].shape[1] for i in range(j))
            for j in range(n_blocks + 1)
        ]
        width = block_x_coords[-1]
        new_shape = (height, width, depth)

        new_regions = []
        new_callbacks = []
        new_vars_dic = {}
        new_image_defaults = {}
        for i, clickable in enumerate(clickables):
            new_callbacks += clickable.callbacks
            new_vars_dic |= clickable.vars_dic
            new_image_defaults |= clickable.image.defaults_dic

            # Checking for interactions between the blocks.
            for j, other_clickable in enumerate(clickables):
                if j > i and other_clickable.has_interaction(clickable):
                    print(
                        "It is likely that you are setting up some "
                        + "interactions between the "
                        + str(i)
                        + "th and "
                        + str(j)
                        + "th passed clickables (counting from 0): "
                        + "they have common keys in their respective "
                        + "'vars_dics | image.defaults_dic'."
                    )

            # Regions processing.
            empty_left_block = np.zeros(
                (height, block_x_coords[i]), dtype="bool"
            )
            empty_right_block = np.zeros(
                (height, width - block_x_coords[i + 1]), dtype="bool"
            )
            for region in clickable.regions:
                translated_region = np.hstack(
                    [empty_left_block, region, empty_right_block]
                )
                new_regions.append(translated_region)

        # new_image callable definition

        def new_image_func(**kwargs):
            for clickable in clickables:
                blocks = [
                    clickable.get_image(kwargs) for clickable in clickables
                ]
            return np.hstack(blocks)

        # new_image definition
        new_image = _CombinableFunc(new_image_func, new_image_defaults)

        return ClickableImage(
            new_image, new_shape, new_regions, new_callbacks, new_vars_dic
        )

    @staticmethod
    def vstack(
        clickables: list[ClickableImage | np.ndarray()],
    ) -> ClickableImage:
        """Stack vertically the input ClickableImages or np.ndarray images.

        Args:
            clickables (list[ClickableImage|np.ndarray]): list of
                ClickableImages or images, in the second case they are treated
                as'constant' ClickableImages.
        Raises:
            ValueError: The clickables are not stackable as demanded,
                due to shape issues. Remember this could include a depth issue.

        Returns:
            ClickableImage: The ClickableImage obtained by stacking
                vertically the input clickables, with possible
                interactions if they share keys of their
                image.defaults_dic | vars_dic
        """
        for i in range(len(clickables)):
            if isinstance(clickables[i], np.ndarray):
                clickables[i] = ClickableImage.constant_to_clickable(
                    clickables[i]
                )
        new_clickables = [clickable.transpose() for clickable in clickables]
        return ClickableImage.hstack(new_clickables).transpose()

    def resize(
        self,
        shape: tuple[int, ...],
        anti_aliasing: bool = False,
        preserve_range: bool = False,
    ) -> ClickableImage:
        """Resize self to input shape.

        Args:
            shape (tuple[int,...]): Desired shape for the output.
            anti_aliasing (bool, optional): argument to be passed in
                the underlying skimage.resize call. Defaults to False.
            preserve_range (bool, optional): argument to be passed in
                the underlying skimage.resize call. Defaults to False.
        """

        def new_image_func(**kwargs):
            return skimage.transform.resize(
                self.image(**kwargs),
                shape,
                anti_aliasing=anti_aliasing,
                preserve_range=preserve_range,
            )

        new_image = _CombinableFunc(new_image_func, self.image.defaults_dic)
        new_shape = shape
        new_regions = [
            skimage.transform.resize(region, shape[:2])
            for region in self.regions
        ]

        return ClickableImage(
            new_image, new_shape, new_regions, self.callbacks, self.vars_dic
        )


# Tests for the ClickableImage class.
# A = np.array([[1, 1], [0, 0]], dtype="bool")
# def image(is_on=True):
#     return is_on & A
# def toggle(dic):
#     dic["is_on"] ^= True


# u = ClickableImage(image, (2, 2), [A], [toggle], {"is_on": True})

# print(u.get_image())
# u.click_action(0)
# print(u.get_image())
# u.click_action(0)
# print(u.get_image())


# v = ClickableImage.hstack([u, u])
# print(v.callbacks)
# print(v.regions)
# print(v.vars_dic)
# v.click_action(1)
# print(v.vars_dic)

# print(v.use())
# print(u.transpose().use())
# print(ClickableImage.vstack([u, u]).use(cmap="gray"))


class ImageToggles(ClickableImage):
    """A subclass of ClickableImage where regions of the image become toggles.

    No new method, a unique original method: __init__ which 'extends' the
    superclass's method."""

    def __init__(
        self,
        image: np.ndarray,
        regions: list[np.ndarray],
        *,
        daltonism=False,
        on_color="green",
        off_color="red",
        default_toggle_value=True,
        true_is_black_pixel=True,
        toggles_name="toggles",
    ):
        """Build self. Extend super().__init__, with distinct arguments.

        Args:
            image (np.ndarray): The "background" image
            regions (list[np.ndarray]): the 'toggle' regions as boolean images.
            daltonism (bool, optional): Option to set a special color mode for
                color blind people. Defaults to False.
            on_color (str, optional): The color of the regions when toggled to
                True. Defaults to "green". This string must be recognized
                by matplotlib.colors.to_rgb.
            off_color (str, optional): The color of the regions when toggled to
                False. Defaults to "red".  This string must be recognized
                by matplotlib.colors.to_rgb.
            default_toggle_value (bool, optional): The common value of all
            toggles as defined in the Clickable image vars_dic. Defaults to True.
            true_is_black_pixel (bool, optional): Decide if the front pixel
                are black, in case image is boolean. Defaults to True.
            toggles_name (str, optional): the key of the toggles values list in
                the ClickableImage's vars_dic. Defaults to "toggles".

        Raises:
            TypeError: "The input regions do not have 'bool' dtype.")

        """
        if all(region.dtype == np.dtype("bool") for region in regions):
            self.regions = regions
        else:
            raise TypeError("The input regions do not have 'bool' dtype.")

        # Defining a clickable image, we first calculate the 'image' attribute.
        if daltonism:
            on_color = "cyan"
            off_color = "magenta"
        on_color = colors.color_to_rgb(on_color)
        off_color = colors.color_to_rgb(off_color)

        my_colors = [off_color, on_color]

        def _displayed_image(**kwargs):
            toggles = kwargs[toggles_name]
            if true_is_black_pixel and image.ndim == 2:
                output = colors.to_rgb(image ^ True)
            else:
                output = colors.to_rgb(image)

            for i, region in enumerate(self.regions):
                output[region] = my_colors[toggles[i]]
            return output

        vars_dic = {toggles_name: [default_toggle_value] * len(regions)}
        displayed_image = _CombinableFunc(_displayed_image, vars_dic)

        def toggle_family(i):
            def f(dic):
                dic[toggles_name] = [
                    dic[toggles_name][j]
                    if j != i
                    else dic[toggles_name][i] ^ True
                    for j in range(len(dic[toggles_name]))
                ]

            return f

        callbacks = [toggle_family(i) for i in range(len(regions))]

        super().__init__(
            displayed_image,
            image.shape[:2] + (3,),
            regions,
            callbacks,
            vars_dic,
        )


class SliceDisplayer(ClickableImage):
    """This subclass is for clickable images that display a slice of text.

    The slice may have variable start and stop.
    """

    def __init__(
        self,
        text: str,
        shape: tuple[int, int],
        *,
        background_color="white",
        slice_varname="slice",
    ):
        """Extends __init__ of superclass, with distinct arguments.

        Args:
            text (str): The text whose slice will be displayed
            shape (tuple[int, int]): The shape of the resulting ClickableImage
            background_color (str, optional): Background color, this string
                must be recognized by matplotlib.colors.to_rgb.
                Defaults to "white".
            slice_varname (str, optional): The name of the variable that
                defines the slice to take. Defaults to "slice".

        """
        defaults_dic = {slice_varname: slice(None, 7)}

        def text_image_(**kwargs):
            return image_annotation.center_text(
                text[kwargs[slice_varname]],
                shape,
                background_color=background_color,
            )
            

        text_image = _CombinableFunc(text_image_, defaults_dic)
        regions = []
        callbacks = []
        super().__init__(
            text_image, shape + (3,), regions, callbacks, defaults_dic
        )


class Button(ClickableImage):
    """The subclass of ClickableImage made of basic rectangular buttons.

    More precisely, the unique clickable region is the whole image.
    """

    def __init__(self, image: np.ndarray, callback: Callable[dict, None]):
        """Button constructor.

        Args:
            image (np.ndarray): The underlying image of the button.
            callback (Callable): The callback triggered by clicking the button.
        """
        if image.ndim == 3 and image.shape[2] == 3:
            image[:, 0] = np.array([0, 0, 0])
            image[:, -1] = np.array([0, 0, 0])
            image[0, :] = np.array([0, 0, 0])
            image[-1, :] = np.array([0, 0, 0])
        elif image.ndim == 2:
            image[:, 0] = 0
            image[:, -1] = 0
            image[0, :] = 0
            image[-1, :] = 0
        else:
            raise ValueError(
                "The image argument is expected to be a "
                + "binary, grayscale or rgb image."
            )

        super().__init__(
            image,
            image.shape,
            [np.ones(image.shape[:2], dtype="bool")],
            [callback],
            vars_dic={},
        )



""" 
+++++++++++++++++++++
The np_gui package
+++++++++++++++++++++

This package implements Graphical User Interfaces from a NumPy user
perspective: every image corresponds to an ndarray.

The core class is \
ClickableImage whose instances' attributes are: 


* a dictionary of parameter values,
* a callable that returns an image of fixed shape,
* regions in that shape defined as boolean images of the same shape 
* and their corresponding callbacks.

When calling the use method of an instance,
the dictionary is passed to the callable, that uses (some of) its values
as arguments and outputs an image which is displayed in a window.

Upon clicking on the distinguished regions, their callbacks are called
and alter the parameter values. The image is then refreshed from this 
input.

When closing the window, the use method returns the dictionary with
its altered values; whence the interaction with the user.

These objects can be stacked vertically and horizontally to compose more
complex ones. A certain number of subclasses are proposed as basic 
building blocks. The stacked blocks may have parameter keywords in common,
which allows for interactions between them.

The code is organized in three submodules as follows.

* **np_clickable_images** contains the definition of the ClickableImage \
class and some subclasses,
* **colors** contains some color management facilities and
* **image_annotation** regards putting text in an image.


"""